/////////////////////////////////////////////////////////////////////////////
//
// wa_viz.cpp : Implementation of CWa_viz
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "wa_viz.h"
#include <stdio.h>

#define SKIN_PATH "D:\\PROGRAM FILES\\WINAMP\\SKINS\\Winamp Modern"
#define INI_PATH "D:\\PROGRAM FILES\\WINAMP\\winamp.ini"

#define IPC_CHECKINITIALIZED 98765

static CWa_viz *vizSingleton;
static char *charBuf;

static WNDPROC origWndProc;
static LONG origWndProcLong;

HWND embedWindowStateFunc(embedWindowState *v)
{
	RECT rect;
	v->me = vizSingleton->GetWindow();
	GetClientRect(v->me, &rect);
	v->r.left = 0;
	v->r.top = 0;
	v->r.right = rect.right;
	v->r.bottom = rect.bottom;
	return v->me;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	char id[256];
	if (msg == WM_WA_IPC) {
		if (wParam == IPC_CHECKINITIALIZED) {
			CWa_viz *viz = (CWa_viz *)lParam;
			viz->InitModule();
			return S_OK;
		}
		switch (lParam) {
			case IPC_GETVERSION:
				return (LRESULT)0x2900;
				break;
			case IPC_GET_EMBEDIF:
				return (LRESULT)embedWindowStateFunc;
				return (LRESULT)vizSingleton->GetWindow();
				break;
			case IPC_SETVISWND:
				// param is hwnd, setting this allows you to receive ID_VIS_NEXT/PREVOUS/RANDOM/FS wm_commands
				break;
			case IPC_GETSKIN:
//				strcpy((char *)wParam, SKIN_PATH);
				break;
			case IPC_ISPLAYING:
				return (LRESULT)1;
				break;
			case IPC_GETINIFILE:
				strcpy(charBuf, INI_PATH);
				return (LRESULT)charBuf;
				break;
			default:
				sprintf(id, "WM_WA_IPC: %u", lParam);
				::MessageBox(NULL, id, "Info", MB_OK);
		}
		return S_OK;
	}
	else if (msg == WM_SIZE) {
		vizSingleton->ResizeModule();
	}
//	return S_OK;
	return CallWindowProc(origWndProc,hWnd,msg,wParam,lParam) ;
}

/////////////////////////////////////////////////////////////////////////////
// CWa_viz::CWa_viz
// Constructor

CWa_viz::CWa_viz() :
m_hwndParent(NULL),
m_hwndVis(NULL),
m_clrForeground(0x0000FF),
m_nPreset(0),
m_hDLL(0),
m_visHdr(0),
m_visMod(0),
m_presetCount(-1),
m_modInitialized(false),
m_configMode(false)
{
	vizSingleton = this;
	origWndProc = NULL;
	origWndProcLong = 0;
}

/////////////////////////////////////////////////////////////////////////////
// CWa_viz::~CWa_viz
// Destructor

CWa_viz::~CWa_viz()
{
	vizSingleton = NULL;
}

/////////////////////////////////////////////////////////////////////////////
// CWa_viz:::FinalConstruct
// Called when an effect is first loaded. Use this function to do one-time
// intializations that could fail (i.e. creating offscreen buffers) instead
// of doing this in the constructor, which cannot return an error.

HRESULT CWa_viz::FinalConstruct()
{
	charBuf = (char *)malloc(1024);
	m_hDLL = LoadLibrary("c:\\program files\\winamp\\plugins\\vis_milk.dll");
	if (!m_hDLL) {
		::MessageBox(NULL, "Could not load DLL", "Error", MB_OK);
		return E_FAIL;
	}

	return S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// CWa_viz:::FinalRelease
// Called when an effect is unloaded. Use this function to free any
// resources allocated in FinalConstruct.

void CWa_viz::FinalRelease()
{
	if (m_hDLL) {
//		FreeLibrary(m_hDLL);
		m_hDLL = NULL;
	}
	free(charBuf);
    ReleaseCore();
}

BOOL CALLBACK EnumChildProc(HWND hwnd,LPARAM lParam)
{
	*(HWND *)lParam = hwnd;
	return FALSE;
}

void CWa_viz::ResizeModule()
{	
	if (!m_hwndVis)
		return;
	RECT rect;
	GetClientRect(m_hwndParent, &rect);
	SetWindowPos(m_hwndVis, HWND_TOP, 0, 0, rect.right, rect.bottom, 0);
}

void CWa_viz::InitModule()
{
	if (m_modInitialized || !m_visMod)
		return;
	if (!(m_visHdr=GetVisHeader()))
		return;
	m_modInitialized = true;
	m_visMod->hwndParent = m_hwndParent;
	m_visMod->hDllInstance = m_hDLL;
	m_visMod->nCh = 2;
	m_visMod->spectrumNch = 2;
	m_visMod->waveformNch = 2;
	m_visMod->sRate = 1000;
	m_visMod->Init(m_visMod);

	m_hwndVis = NULL;
	EnumChildWindows(m_hwndParent, EnumChildProc, (LPARAM)&m_hwndVis);
	ResizeModule();
}

winampVisHeader *CWa_viz::GetVisHeader()
{
	if (m_visHdr)
		return m_visHdr;

	if (!m_hDLL)
		return NULL;

	winampVisGetHeaderType winampVisGetHeader;
	winampVisGetHeader = (winampVisGetHeaderType)GetProcAddress(m_hDLL,
                                           "winampVisGetHeader");
	winampVisHeader *hdr = winampVisGetHeader();
	return hdr;
}

BOOL CWa_viz::LoadPresets()
{
	if (m_presetCount >= 0)
		return true;

	if (!(m_visHdr = GetVisHeader())) {
		::MessageBox(NULL, "Could not load header", "WinAmp vis plugin", MB_OK);
		return false;
	}

	m_presetCount = 0;
	while (m_visHdr->getModule((int)m_presetCount)) {
		m_presetCount++; 
	}
	return true;
}

//////////////////////////////////////////////////////////////////////////////
// CWa_viz::Render
// Called when an effect should render itself to the screen.
//////////////////////////////////////////////////////////////////////////////
STDMETHODIMP CWa_viz::Render(TimedLevel *pLevels, HDC hdc, RECT *prc)
{
	if (m_hwndParent == NULL) {
	  ::FillRect(hdc, prc, CreateSolidBrush(RGB(0,0,0)));
  	  ::DrawText(hdc, "Can't run windowless", 20, prc, DT_CENTER | DT_NOCLIP);
	  return S_OK;
	}

	if (!(m_visHdr=GetVisHeader()))
		return E_FAIL;

	if (!m_modInitialized) {
		SendMessage(m_hwndParent, WM_WA_IPC, IPC_CHECKINITIALIZED, (LPARAM)this);
	}
	else if (m_configMode) {
		return S_OK;
	}
	else {
		winampVisModule *mod = m_visMod;
		if (!mod)
			return E_FAIL;

		for (int i=0; i<512; i++) {
			mod->spectrumData[0][i] = pLevels->frequency[0][i*2];
			mod->waveformData[0][i] = pLevels->waveform[0][i*2]^128;
			mod->spectrumData[1][i] = pLevels->frequency[1][i*2];
			mod->waveformData[1][i] = pLevels->waveform[1][i*2]^128;
		}
/*
		for (int i=512;i<576;i++) {
			mod->spectrumData[0][i] = 0;
			mod->waveformData[0][i] = 0;
			mod->spectrumData[1][i] = 0;
			mod->waveformData[1][i] = 0;
		}
		*/
		mod->Render(mod);
	}

    return S_OK;
}

//////////////////////////////////////////////////////////////////////////////
// CWa_viz::MediaInfo
// Everytime new media is loaded, this method is called to pass the
// number of channels (mono/stereo), the sample rate of the media, and the
// title of the media
//////////////////////////////////////////////////////////////////////////////
STDMETHODIMP CWa_viz::MediaInfo(LONG lChannelCount, LONG lSampleRate, BSTR bstrTitle )
{
    return S_OK;
}


//////////////////////////////////////////////////////////////////////////////
// CWa_viz::GetCapabilities
// Returns the capabilities of this effect. Flags that can be returned are:
//	EFFECT_CANGOFULLSCREEN		-- effect supports full-screen rendering
//	EFFECT_HASPROPERTYPAGE		-- effect supports a property page
//////////////////////////////////////////////////////////////////////////////
STDMETHODIMP CWa_viz::GetCapabilities(DWORD * pdwCapabilities)
{
    if (NULL == pdwCapabilities)
    {
        return E_POINTER;
    }

    *pdwCapabilities = EFFECT_HASPROPERTYPAGE;
    return S_OK;
}

//////////////////////////////////////////////////////////////////////////////
// CWa_viz::GetTitle
// Invoked when a host wants to obtain the title of the effect
//////////////////////////////////////////////////////////////////////////////
STDMETHODIMP CWa_viz::GetTitle(BSTR* bstrTitle)
{
	::MessageBox(NULL, "GetTitle", "Info", MB_OK);
    USES_CONVERSION;

    if (NULL == bstrTitle)
    {
        return E_POINTER;
    }

	if(!(m_visHdr = GetVisHeader()))
		return E_FAIL;

	CComBSTR bstrTemp(m_visHdr->description);
        
    if ((!bstrTemp) || (0 == bstrTemp.Length()))
    {
        return E_FAIL;
    }

	bstrTemp.CopyTo(bstrTitle);

    return S_OK;
}

//////////////////////////////////////////////////////////////////////////////
// CWa_viz::GetPresetTitle
// Invoked when a host wants to obtain the title of the given preset
//////////////////////////////////////////////////////////////////////////////
STDMETHODIMP CWa_viz::GetPresetTitle(LONG nPreset, BSTR *bstrPresetTitle)
{
    USES_CONVERSION;

    if (NULL == bstrPresetTitle)
    {
        return E_POINTER;
    }

	if (!LoadPresets())
		return E_FAIL;

    if ((nPreset < 0) || (nPreset >= m_presetCount))
    {
        return E_INVALIDARG;
    }

	winampVisModule *visMod = m_visHdr->getModule((int)nPreset);
	if (!visMod) {
		::MessageBox(NULL, "Could not load preset", "WinAmp vis plugin", MB_OK);
		return E_FAIL;
	}
    CComBSTR bstrTemp(visMod->description);

    if ((!bstrTemp) || (0 == bstrTemp.Length()))
    {
        return E_FAIL;
    }

	bstrTemp.CopyTo(bstrPresetTitle);

    return S_OK;
}

//////////////////////////////////////////////////////////////////////////////
// CWa_viz::GetPresetCount
// Invoked when a host wants to obtain the number of supported presets
//////////////////////////////////////////////////////////////////////////////
STDMETHODIMP CWa_viz::GetPresetCount(LONG *pnPresetCount)
{
    if (NULL == pnPresetCount)
    {
        return E_POINTER;
    }

	if (!LoadPresets())
		return E_FAIL;

    *pnPresetCount = m_presetCount;

    return S_OK;
}

//////////////////////////////////////////////////////////////////////////////
// CWa_viz::SetCurrentPreset
// Invoked when a host wants to change the index of the current preset
//////////////////////////////////////////////////////////////////////////////
STDMETHODIMP CWa_viz::SetCurrentPreset(LONG nPreset)
{
    if ((nPreset < 0) || (nPreset >= PRESET_COUNT))
    {
        return E_INVALIDARG;
    }

    m_nPreset = nPreset;

    return S_OK;
}

//////////////////////////////////////////////////////////////////////////////
// CWa_viz::GetCurrentPreset
// Invoked when a host wants to obtain the index of the current preset
//////////////////////////////////////////////////////////////////////////////
STDMETHODIMP CWa_viz::GetCurrentPreset(LONG *pnPreset)
{
    if (NULL == pnPreset)
    {
        return E_POINTER;
    }

    *pnPreset = m_nPreset;

    return S_OK;
}

//////////////////////////////////////////////////////////////////////////////
// CWa_viz::SetCore
// Set WMP core interface
//////////////////////////////////////////////////////////////////////////////
STDMETHODIMP CWa_viz::SetCore(IWMPCore * pCore)
{
    HRESULT hr = S_OK;

    // release any existing WMP core interfaces
    ReleaseCore();

    // If we get passed a NULL core, this  means
    // that the plugin is being shutdown.

    if (pCore == NULL)
    {
        return S_OK;
    }

    m_spCore = pCore;

    return hr;
}

//////////////////////////////////////////////////////////////////////////////
// CWa_viz::Create
// Invoked when the visualization should be initialized.
//
// If hwndParent != NULL, RenderWindowed() will be called and the visualization
// should draw into the window specified by hwndParent. This will be the
// behavior when the visualization is hosted in a window.
//
// If hwndParent == NULL, Render() will be called and the visualization
// should draw into the DC passed to Render(). This will be the behavior when
// the visualization is hosted windowless (like in a skin for example).
//////////////////////////////////////////////////////////////////////////////
STDMETHODIMP CWa_viz::Create(HWND hwndParent)
{
	if (!origWndProc) {
		origWndProc = (WNDPROC)GetWindowLong(hwndParent, GWL_WNDPROC);
		origWndProcLong = SetWindowLong(hwndParent, GWL_WNDPROC, (LONG)WndProc);
	}
	m_modInitialized = false;
	m_hwndParent = hwndParent;
	if (!(m_visHdr=GetVisHeader()))
		return E_FAIL;
	m_visMod = m_visHdr->getModule(m_nPreset);
	return S_OK;
}

STDMETHODIMP CWa_viz::DisplayPropertyPage(HWND hwndOwner) 
{ 
	if (!(m_visHdr=GetVisHeader()))
		return E_FAIL;
	m_visMod = m_visHdr->getModule(m_nPreset);
	m_visMod->hwndParent = hwndOwner;
	m_visMod->hDllInstance = m_hDLL;
	m_visMod->Config(m_visMod);
	return S_OK; 
};
//////////////////////////////////////////////////////////////////////////////
// CWa_viz::Destroy
// Invoked when the visualization should be released.
//
// Any resources allocated for rendering should be released.
//////////////////////////////////////////////////////////////////////////////
STDMETHODIMP CWa_viz::Destroy()
{
	if (!(m_visHdr=GetVisHeader()))
		return E_FAIL;
	if (m_visMod) {
		m_visMod->Quit(m_visMod);
		m_visMod = NULL;
	}
	m_modInitialized = false;
	SetWindowLong(m_hwndParent,GWL_WNDPROC,origWndProcLong);
	m_hwndVis = NULL;
//    m_hwndParent = NULL;

    return S_OK;
}

//////////////////////////////////////////////////////////////////////////////
// CWa_viz::NotifyNewMedia
// Invoked when a new media stream begins playing.
//
// The visualization can inspect this object for properties (like name or artist)
// that might be interesting for visualization.
//////////////////////////////////////////////////////////////////////////////
STDMETHODIMP CWa_viz::NotifyNewMedia(IWMPMedia *pMedia)
{
    return S_OK;
}

//////////////////////////////////////////////////////////////////////////////
// CWa_viz::OnWindowMessage
// Window messages sent to the parent window.
//////////////////////////////////////////////////////////////////////////////
STDMETHODIMP CWa_viz::OnWindowMessage(UINT msg, WPARAM wParam, LPARAM lParam, LRESULT *lResult )
{
	char id[256];
	sprintf(id, "MSG: %u", msg);
	if (msg == WM_WA_IPC) {
		if (wParam == IPC_CHECKINITIALIZED) {
			CWa_viz *viz = (CWa_viz *)lParam;
			viz->InitModule();
			return S_OK;
		}

	        // These are IPC messages sent by the viz dll that need to be handled
                //  Normally, these would be handled by WinAmp, so we have to trick the DLL into
                //  thinking we are winamp.
		switch (lParam) {
			case IPC_GETVERSION:
				*lResult = (LRESULT)0x5090;
				break;
			case IPC_GET_EMBEDIF:
				*lResult = (LRESULT)embedWindowStateFunc;
				break;
			case IPC_SETVISWND:
				// param is hwnd, setting this allows you to receive ID_VIS_NEXT/PREVOUS/RANDOM/FS wm_commands
				break;
			case IPC_GETSKIN:
//				strcpy((char *)wParam, SKIN_PATH);
				break;
			case IPC_GETINIFILE:
//				strcpy(charBuf, INI_PATH);
//				return (LRESULT)charBuf;
				break;
			default:
				sprintf(id, "WM_WA_IPC: %u", lParam);
				::MessageBox(NULL, id, "Info", MB_OK);
		}
		return S_OK;
	}
	else if (msg == WM_SIZE) {
		::MessageBox(NULL, "OnWSIZE", "Info", MB_OK);
	}
	return S_FALSE;
}

//////////////////////////////////////////////////////////////////////////////
// CWa_viz::RenderWindowed
// Called when an effect should render itself to the screen.
//
// The fRequiredRender flag specifies if an update is required, otherwise the
// update is optional. This allows visualizations that are fairly static (for example,
// album art visualizations) to only render when the parent window requires it,
// instead of n times a second for dynamic visualizations.
//////////////////////////////////////////////////////////////////////////////
STDMETHODIMP CWa_viz::RenderWindowed(TimedLevel *pLevels, BOOL fRequiredRender )
{
    // NULL parent window should not happen 
    if (NULL == m_hwndParent)
    {
        return E_UNEXPECTED;
    }

    // At this point the visualization should draw directly into the parent
    // window. This sample just calls windowless render for simplicity.

    HDC hdc = ::GetDC(m_hwndParent);

    if (NULL == hdc)
    {
        return E_FAIL;
    }

    RECT rParent = { 0 };
    ::GetClientRect(m_hwndParent, &rParent);

    Render(pLevels, hdc, &rParent);
    
    ::ReleaseDC(m_hwndParent, hdc);

    return S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// CWa_viz::ReleaseCore
// Release WMP core interfaces
//////////////////////////////////////////////////////////////////////////////
void CWa_viz::ReleaseCore()
{
    if (m_spCore)
    {
        m_spCore = NULL;
    }
}

//////////////////////////////////////////////////////////////////////////////
// CWa_viz::get_foregroundColor
// Property get to retrieve the foregroundColor prop via the public interface.
//////////////////////////////////////////////////////////////////////////////
STDMETHODIMP CWa_viz::get_foregroundColor(BSTR *pVal)
{
	return ColorToWz( pVal, m_clrForeground);
}


//////////////////////////////////////////////////////////////////////////////
// CWa_viz::put_foregroundColor
// Property put to set the foregroundColor prop via the public interface.
//////////////////////////////////////////////////////////////////////////////
STDMETHODIMP CWa_viz::put_foregroundColor(BSTR newVal)
{
	return WzToColor(newVal, &m_clrForeground);
}


//////////////////////////////////////////////////////////////////////////////
// CWa_viz::WzToColor
// Helper function used to convert a string into a COLORREF.
//////////////////////////////////////////////////////////////////////////////
HRESULT CWa_viz::WzToColor(const WCHAR *pwszColor, COLORREF *pcrColor)
{ 
    if (NULL == pwszColor)
    {
        //NULL color string passed in
        return E_POINTER;
    }

    if (0 == lstrlenW(pwszColor))
    {
        //Empty color string passed in
        return E_INVALIDARG;
    }

    if (NULL == pcrColor)
    {
        //NULL output color DWORD passed in
        return E_POINTER;
    }
    
    if (lstrlenW(pwszColor) != 7)
    {
        //hex color string is not of the correct length
        return E_INVALIDARG;
    }

    DWORD dwRet = 0;
    for (int i = 1; i < 7; i++)
    {
        // shift dwRet by 4
        dwRet <<= 4;
        // and add in the value of this string

        if ((pwszColor[i] >= L'0') && (pwszColor[i] <= L'9'))
        {
            dwRet += pwszColor[i] - '0';
        }
        else if ((pwszColor[i] >= L'A') && (pwszColor[i] <= L'F'))
        {
            dwRet += 10 + (pwszColor[i] - L'A');
        }
        else if ((pwszColor[i] >= L'a') && (pwszColor[i] <= L'f'))
        {
            dwRet += 10 + (pwszColor[i] - L'a');
        }
        else
        {
           //Invalid hex digit in color string
            return E_INVALIDARG;
        }
    }

    *pcrColor = SwapBytes(dwRet);

    return S_OK;
} 


//////////////////////////////////////////////////////////////////////////////
// CWa_viz::ColorToWz
// Helper function used to convert a COLORREF to a BSTR.
//////////////////////////////////////////////////////////////////////////////
HRESULT CWa_viz::ColorToWz( BSTR* pbstrColor, COLORREF crColor)
{
    _ASSERT( NULL != pbstrColor );
    _ASSERT( (crColor & 0x00FFFFFF) == crColor );

    *pbstrColor = NULL;

    WCHAR wsz[8];
    HRESULT hr  = S_OK;

    wsprintfW( wsz, L"#%06X", SwapBytes(crColor) );
    
    *pbstrColor = ::SysAllocString( wsz );

    if (!pbstrColor)
    {
        hr = E_FAIL;
    }

    return hr;
}


//////////////////////////////////////////////////////////////////////////////
// CWa_viz::SwapBytes
// Used to convert between a DWORD and COLORREF.  Simply swaps the lowest 
// and 3rd order bytes.
//////////////////////////////////////////////////////////////////////////////
inline DWORD CWa_viz::SwapBytes(DWORD dwRet)
{
    return ((dwRet & 0x0000FF00) | ((dwRet & 0x00FF0000) >> 16) | ((dwRet & 0x000000FF) << 16));
}

